@@ -1,6 +1,10 @@
 
 `%||%` <- function(l, r) if (is.null(l)) r else l
 
+new_class <- function(class_name, ...) {
+  structure(as.environment(list(...)), class = class_name)
+}
+
 make_space <- function(len) {
   strrep(" ", len)
 }
@@ -39,11 +43,11 @@ apply_style <- function(text, style, bg = FALSE) {
   if (is.function(style)) {
     style(text)
   } else if (is.character(style)) {
-    crayon::make_style(style, bg = bg)(text)
+    make_ansi_style(style, bg = bg)(text)
   } else if (is.null(style)) {
     text
   } else {
-    stop("Not a colour name or crayon style", call. = FALSE)
+    stop("Not a colour name or ANSI style function", call. = FALSE)
   }
 }
 
@@ -64,7 +68,7 @@ ruler <- function(width = console_width()) {
   y <- rep("-", length(x))
 
   y[x %% 5 == 0] <- "+"
-  y[x %% 10 == 0] <- crayon::bold(as.character((x[x %% 10 == 0] %/% 10) %% 10))
+  y[x %% 10 == 0] <- style_bold(as.character((x[x %% 10 == 0] %/% 10) %% 10))
 
   cat(y, "\n", sep = "")
   cat(x %% 10, "\n", sep = "")
@@ -105,10 +109,41 @@ new_uuid <- (function() {
   cnt <- 0
   function() {
     cnt <<- cnt + 1
-    paste0("cli", cnt)
+    paste0("cli-", clienv$pid, "-", cnt)
   }
 })()
 
 na.omit <- function(x) {
   if (is.atomic(x)) x[!is.na(x)] else x
 }
+
+last <- function(x) {
+  tail(x, 1)[[1]]
+}
+
+str_tail <- function(x) {
+  substr(x, 2, nchar(x))
+}
+
+push <- function(l, el, name = NULL) {
+  c(l, structure(list(el), names = name))
+}
+
+try_silently <- function(expr) {
+  suppressWarnings(tryCatch(expr, error = function(x) x))
+}
+
+random_id <- function() {
+  paste(sample(c(letters, LETTERS, 0:9), 7, replace = TRUE), collapse = "")
+}
+
+str_trim <- function(x) {
+  sub("^\\s+", "", sub("\\s+$", "", x))
+}
+
+has_asciicast_support <- function() {
+ tryCatch({
+   asNamespace("asciicast")$is_recording_supported() &&
+     asNamespace("asciicast")$is_svg_supported()
+ }, error = function(e) FALSE)
+}
